/*---------------------------------------------------------------------------

    File: WBTREE.C
    Author: Will Hobday

    Description: module to implement a weight balanced tree

---------------------------------------------------------------------------*/

#include <malloc.h>
#include <string.h>
#include <stdio.h>
#include "wbtree.h"

#define ALPHA 0.2928932188  /* 1 - (2/SQR(2)) */

WBT_ptr WBT_root;

/*---------------------------- Internal Funtctions ------------------------*/

WBT_ptr _WBT_alloc_node(SAMPLE_data_type data)
/* --- --- ---
    _alloc_node(internal)

    Parms:      data - data to intialize new node
    Returns:    pointer to allocated node

    Desc:  Internal func that uses malloc to allocated a block of memory for
           a new node
*/
{
    WBT_ptr p;

    if ((p = (WBT_ptr)malloc(sizeof(struct WBT_node))) == NULL)
    {
        fprintf(stderr,"_WBT_alloc_node: out of memory\n");
        exit(1);
    }

    p->data = data;
    p->right = p->left = NULL;
    p->weight = 2;

    return p;
}

#define weight(p) (((p == NULL) ? 1 : p->weight))
#define balance_factor(p) (((float)weight(p->left) / (float)weight(p)))

WBT_ptr _WBT_rotate_left(WBT_ptr p)
/* --- --- ---
    _WBT_rotate_left/_WBT_rotate_right(internal)

    Parms:      p - root of tree to rotate
    Returns:    pointer to rotated tree

    Desc:  Function to preform standard left/right tree rotation.  Adjusts
           pointers and recalculates node weight.
*/
{
    WBT_ptr q;

    q = p->left;
    p->left = q->right;
    q->right = p;

    p->weight = weight(p->left) + weight(p->right);
    q->weight = weight(q->left) + weight(q->right);

    return q;
}

WBT_ptr _WBT_rotate_right(WBT_ptr p)
{
    WBT_ptr q;

    q = p->right;
    p->right = q->left;
    q->left = p;

    p->weight = weight(p->left) + weight(p->right);
    q->weight = weight(q->left) + weight(q->right);

    return q;
}

WBT_ptr _WBT_balance(WBT_ptr p, WBT_ptr parent)
/* --- --- ---
    _WBT_balance(internal)

    Parms:      p - root of tree to rotate, parent - parent of node p
    Returns:    pointer to balanced tree.

    Desc:  Function to balance a given tree.  The tree is balanced by
           using the rotate functions defined above.
*/
{
    WBT_ptr q;

    p->weight = weight(p->left) + weight(p->right);

    if (balance_factor(p) > (1 - ALPHA))
    {
        if (balance_factor(p->left) <= (ALPHA/(1 - ALPHA)))
            p->left = _WBT_rotate_right(p->left);
        q = _WBT_rotate_left(p);
        goto rotate;
    }

    if (balance_factor(p) < ALPHA)
    {
        if (balance_factor(p->right) >= ((1 - 2*ALPHA)/(1 - ALPHA)))
            p->right = _WBT_rotate_left(p->right);
        q = _WBT_rotate_right(p);
        goto rotate;
    }
    return p;
rotate:
    if (parent != NULL)
        if ( parent->right == p)
            parent->right = q;
        else
            parent->left = q;
    return q;
}



/* ---------------------------- External Funtctions ------------------------*/



void WBT_kill_tree(WBT_ptr p)
/* --- --- ---
    WBT_kill_tree(external)

    Parms:      p - root of tree to chop down

    Desc:  Function that deletes an entire tree and free its allocated mem.
*/
{
    if (p != NULL)
    {
        WBT_kill_tree(p->left);
        WBT_kill_tree(p->right);
        free(p);
    }
}

void WBT_traverse(WBT_ptr p)
/* --- --- ---
    WBT_traverse(external)

    Parms:      p - root of tree to traverse

    Desc:  Function that does an inorder traversal of give tree
*/
{
    if (p != NULL)
    {
        WBT_traverse(p->left);
        printf("p: %p\n",p);
        WBT_traverse(p->right);
    }
}

WBT_ptr WBT_search(WBT_ptr p,SAMPLE_data_type key)
/* --- --- ---
    WBT_search(external)

    Parms:      p - root of tree to search, key - key to search for
    Returns:    pointer to node with given key if found, NULL otherwise

    Desc:  Non-recursive binary search.  Returns pointer to node if
           found, NULL otherwise.
*/
{
    while (p != NULL)
    {
        nv++;
        if (key < p->data)
            p = p->left;
        else
        if (key > p->data)
            p = p->right;
        else
            return p;
    }

    return NULL;
}

WBT_ptr WBT_insert(WBT_ptr p,WBT_ptr parent,SAMPLE_data_type data)
/* --- --- ---
    WBT_insert(external)

    Parms:      p - root of tree to insert, parent - parent of node p,
                data - data to insert
    Returns:    pointer to tree with data inserted

    Desc:  Function to insert data into tree.  Uses simple recursive descent
           routine.  Request ignored if data exists in tree.
*/
{
    if (p == NULL)
        return _WBT_alloc_node(data);
    else
    {
        nv++;
        if (data < p->data)
            p->left = WBT_insert(p->left,p,data);
        else
        if (data > p->data)
            p->right = WBT_insert(p->right,p,data);
        else
            return p;

        return _WBT_balance(p,parent);
    }
}

WBT_ptr WBT_delete(WBT_ptr p,WBT_ptr parent,SAMPLE_data_type key)
/* --- --- ---
    WBT_delete(external)

    Parms:      p - root of tree to delete from, parent - parent of node p,
                key - key of node to delete
    Returns:    pointer to tree with data delete

    Desc:  Function to delete data from tree.  Uses simple recursive descent
           routine.  Returns NULL if data doesn't exist.
*/
{
    WBT_ptr q;

    if (p == NULL)
        return NULL;
    else
    {
        nv++;
        if (key < p->data)
            p->left = WBT_delete(p->left,p,key);
        else
        if (key > p->data)
            p->right = WBT_delete(p->right,p,key);
        else
        if (p->left && p->right)
        {
            q = p->left;
            while (q->right != NULL) q = q->right;
            p->data = q->data;
            p->left = WBT_delete(p->left,p,p->data);
        }
        else
        {
            if (p->left == NULL)
                q = p->right;
            if (p->right == NULL)
                q = p->left;
            free(p);

            return q;
        }

        return _WBT_balance(p,parent);
    }
}
